import click
import os
import subprocess
from stak.util import Config
from stak.util import Device
from stak.util import Util
from distutils import spawn

def push_to_device( project, attributes ):
  click.echo( "Pushing project " + click.style( project, bold=True, fg="blue") )
  click.echo( "target: " + click.style( attributes["target"], bold=True, fg="green") )

  # find ssh and rsync executables
  ssh = spawn.find_executable("ssh")
  rsync = spawn.find_executable("rsync")

  # store device address info( user@hostname ) in a variable
  device_address = attributes["device"]["user"]+"@"+attributes["device"]["address"]
  ssh_identity_arg = "'" + ssh+" -o CheckHostIP=no -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no -i " + attributes["device"]["key-file"]+ "'"

  # create the rsync command by joining a list of the arguments, separated by a space

  run_ssh_command( project, attributes, "mkdir -p "+project+"/build" )
  rsync_comand = " ".join([
      rsync,
      "-avz",
      "-e",
      ssh_identity_arg,
      "--delete",
      attributes["output-directory"]+"/",
      device_address+":"+project+"/build"
    ])

  # run the rsync command
  subprocess.Popen(rsync_comand,
    stdout=subprocess.PIPE,
    stderr=subprocess.PIPE,
    shell=True).stdout.readline()

def run_on_device( project, attributes ):
  click.echo( "Running project " + click.style( project, bold=True, fg="blue") )
  click.echo( "target: " + click.style( attributes["target"], bold=True, fg="green") )

  # find ssh executable
  ssh = spawn.find_executable("ssh")

  # store device address info( user@hostname ) in a variable
  device_address = attributes["device"]["user"]+"@"+attributes["device"]["address"]
  click.echo("Running")
  with open("stak.log","wb") as log, open("stak-errors.log","wb") as err:
      subprocess.Popen([ ssh,"-o","CheckHostIP=no","-o", "UserKnownHostsFile=/dev/null", "-o", "StrictHostKeyChecking no", "-i", attributes["device"]["key-file"], device_address, "'"+project+"/" + attributes["executable"]+ "'"], stdout=log, stderr=err)

def is_project( directory ):
  if( os.path.exists( directory ) and
      os.path.isdir ( directory ) and
      os.path.exists( directory + "/stak.yml" ) and
      os.path.isfile( directory + "/stak.yml" ) ):
    return True;
  else:
    return False

def get_project_run_attributes( project, target ):
  
  # get attributes from config
  project_config = Config.get_attributes( [
    [ "default-device", ""],
    [ "executable", project+".sh"],
    [ "output-directory", "build"],
    [ "language", "python"],
    ], Config.local_data)

  if not project_config:
    click.echo("project_config is None")
    return

  # get device attributes from config
  project_config["device"] = Device.get_attributes(
    [
      ["address", ""],
      ["user", "root"],
      ["key-file", ""]
    ],
    project_config["default-device"],
    Config.local_data)

  project_config["target"] = target
  if not project_config["device"]:
    click.echo("device is None")
    return

  return project_config

def create( project, example_options ):

  config = {}
  if "create" in example_options and example_options['create']:
    if "language" in example_options:
      if "python" in example_options["language"]:
        Util.call_and_return("git","clone","git@github.com:NextThingCo/stak-example-python.git", project )
      else:
        os.mkdir( project )
        open( project + "/stak.yml" , 'w')
      config['language'] = example_options['language']
  Config.save( project, config )

def build( project, target ):
  if is_project( project ):
    click.echo( "Building project " + click.style( os.path.basename( project ), bold=True, fg="blue") )
    click.echo( "target: " + click.style( target, bold=True, fg="green") )
    # make sure we have all of the attributes requried to deploy and run the project on a device
    proj_attr = get_project_run_attributes( project, target )

    if not proj_attr:
      click.echo("proj_attr is None")
      return
    if "python" in proj_attr["language"]:
      click.echo("No need to build a python project")
    elif "c++" in proj_attr["language"]:
        Util.call_and_return( "cmake","..", directory="./build")
        Util.call_and_return( "make", directory="./build")
  else:
    click.echo( "Project '" + click.style( os.path.basename( project ), bold=True, fg="blue") +"' not found!")

def ssh( project, target ):
  # verify that we are trying to run a valid project
  if is_project( project ):

    # make sure we have all of the attributes requried to deploy and run the project on a device
    proj_attr = get_project_run_attributes( project, target )

    if not proj_attr:
      click.echo("proj_attr is None")
      return

    project_name = os.path.basename( project )

    # find ssh executable
    ssh = spawn.find_executable("ssh")

    # store device address info( user@hostname ) in a variable
    device_address = proj_attr["device"]["user"]+"@"+proj_attr["device"]["address"]
    subprocess.call([
        ssh,
        "-o","CheckHostIP=no",
        "-o", "UserKnownHostsFile=/dev/null",
        "-o", "StrictHostKeyChecking no",
        "-i", proj_attr["device"]["key-file"],
        device_address
      ])
  else:
    click.echo( "Project '" + click.style( os.path.basename( project ), bold=True, fg="blue") +"' not found!")

def run_ssh_command(project, attributes, command):

  # store device address info( user@hostname ) in a variable
  device_address = attributes["device"]["user"]+"@"+attributes["device"]["address"]

  # find ssh executable
  ssh = spawn.find_executable("ssh")
  return subprocess.Popen([
      ssh,
      "-o","CheckHostIP=no",
      "-o", "UserKnownHostsFile=/dev/null",
      "-o", "StrictHostKeyChecking=no",
      "-i", attributes["device"]["key-file"],
      device_address,
      "%s" % (command)
    ], stdout=subprocess.PIPE, stderr=subprocess.PIPE
  ).stderr.readlines()
  # return " ".join([
  #     ssh,
  #     "-o","CheckHostIP=no",
  #     "-o", "UserKnownHostsFile=/dev/null",
  #     "-o", "StrictHostKeyChecking=no",
  #     "-i", attributes["device"]["key-file"],
  #     device_address,
  #     "'%s'" % (command)
  #   ])


def run( project, target ):

  # verify that we are trying to run a valid project
  if is_project( project ):

    # make sure we have all of the attributes requried to deploy and run the project on a device
    proj_attr = get_project_run_attributes( project, target )

    if not proj_attr:
      click.echo("proj_attr is None")
      return

    project_name = os.path.basename( project )

    push_to_device( project_name, proj_attr )
    run_on_device( project_name, proj_attr )
  else:
    click.echo( "Project '" + click.style( os.path.basename( project ), bold=True, fg="blue") +"' not found!")

def deploy( project, target ):
  if is_project( project ):
    # make sure we have all of the attributes requried to deploy and run the project on a device
    proj_attr = get_project_run_attributes( project, target )

    if not proj_attr:
      click.echo("proj_attr is None")
      return

    project_name = os.path.basename( project )

    Util.call_and_return("honcho", "export", "-f", os.getcwd()+"/build/Procfile", "-d", "/root/" + project_name + "/build", "-l", "/root/" + project_name + "/build", "-a", "stak", "supervisord", os.getcwd()+"/build" )
    
    run_ssh_command( project_name, proj_attr, "virtualenv "+project_name )

    push_to_device( project_name, proj_attr )

    run_ssh_command( project_name, proj_attr, "/root/" + project_name + "/bin/pip install -r /root/"+project_name+"/build/requirements.txt" )    
    
    run_ssh_command( project_name, proj_attr, "/usr/bin/supervisorctl stop all" )
    run_ssh_command( project_name, proj_attr, "mv "+project_name+"/build/stak.conf /etc/supervisor/conf.d/" )
    run_ssh_command( project_name, proj_attr, "/usr/bin/supervisorctl update" )
    run_ssh_command( project_name, proj_attr, "/usr/bin/supervisorctl start all" )
  else:
    click.echo( "Project '" + click.style( os.path.basename( project ), bold=True, fg="blue") +"' not found!")


import click
import subprocess

def get( config, name, parameter ):
  if( "devices" in config and
      name in config["devices"] and
      parameter in config["devices"][name] ):
    return config["devices"][name][parameter]
  else:
    return None

def set( config, name, parameter, value ):
  if( "devices" in config and
      name in config["devices"] and
      parameter in config["devices"][name] ):
    config["devices"][name][parameter] = value

def count( config ):
  if not "devices" in config:
    return 0

  if not config["devices"]:
    return 0

  return len( config["devices"] )

def exists( config, name ):
  if not config:
    return False
  if not "devices" in config:
    return False
  if not config["devices"]:
    return False
  if not name in config["devices"]:
    return False
  return True


def add( config, name, device ):
  if( "devices" in config and
      not name in config["devices"] ):
    config["devices"][name] = device
  return config

def remove( config, name ):
  if( "devices" in config and
      name in config["devices"] ):
    del config["devices"][name]
  return config

def get_attributes( attributes, device, config):
  return_attr = {}

  if not exists( config, device ):
    click.echo(device + " not found")
    click.echo(config)
    return None

  for attr in attributes:
    t_attr = get( config, device, attr[0] )
    if attr[0]:
      return_attr[ attr[0] ] = t_attr
    else:
      click.echo( attr[0]+" not defined for in device config for '"+ device+"'" )
      return_attr[ attr[0] ] = attr[1]
  return return_attr


import click
import yaml
import os, sys
import re
import subprocess
import pwd, grp
from stak.util import DD

# displays a list of data.
# you can control how the data is output by 
def display_list( string, list , transform = ( lambda index,value: click.style( value, bold=True, fg="cyan" ) ) ):
  # list devices
    click.secho( string , bold=True)
    for index, value in enumerate( list ):
      click.echo( "  " + str(index) + ") " + transform( index, value ) )

# calls a shell command and returns the output
def call_and_return( *args, **kwargs ):
  if "directory" in kwargs:
    return subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=kwargs["directory"]).stdout.readlines()
  else:
    return subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE).stdout.readlines()


# utility function for searching a list for a regex.
# you can also pass a lambda in to control how the data
# is transformed before being added to the resulting list
def find_list_items_containing( needle, haystack, transform=( lambda m: { m.group(1), m.group(2) } ) ):
  output = {}
  for item in haystack:
    matches = re.search( needle, item )
    if matches:
      output.update( transform( matches ) )
  if len(output) == 0:
    return None
  return output


def cp( src, dest ):
  call_and_return( "cp",src, dest )

def create_folders_if_nonexistent( *folders ):
  for folder in folders:
    folder_expanded = os.path.expanduser( folder )
    if not os.path.exists( folder_expanded ):
      os.mkdir( folder_expanded )


def dd( src, dst ):
  # defaults
  iput  = DD.FileIo('/dev/stdin', 'r')
  oput  = DD.FileIo('/dev/stdout', 'a')
  ts    = None
  bs    = 2048000
  count = None
  skip  = None
  quiet = False

  if os.path.isfile(src):
    iput = DD.FileIo(src, 'r')
  else:
    if src.find('http'):
      iput = DD.HttpIo( src )
    elif src.find('ftp'):
      iput = DD.FtpIo ( src )

  oput = DD.FileIo(dst, 'w+')
  
  obj = DD.Transfer(iput, oput, ts, bs, count, skip, quiet)
  obj.start()

def raise_privileges( arguments ):
  args = ['sudo', sys.executable] + sys.argv + arguments + [os.environ]
  # the next line replaces the currently-running process with the sudo
  os.execlpe('sudo', *args)

#Throws OSError exception (it will be thrown when the process is not allowed
#to switch its effective UID or GID):
def drop_privileges():
    if os.getuid() != 0:
        # We're not root so, like, whatever dude
        return

    # Get the uid/gid from the name
    user_name = os.getenv("SUDO_USER")
    pwnam = pwd.getpwnam(user_name)

    # Remove group privileges
    os.setgroups([])

    # Try setting the new uid/gid
    os.setgid(pwnam.pw_gid)
    os.setuid(pwnam.pw_uid)

    #Ensure a reasonable umask
    old_umask = os.umask(0o22)


def get_http_header( url ):
  headers = call_and_return( "curl", "-I", url )
  info = find_list_items_containing("^(.*)\:\s(.*)$", headers, ( lambda m: { m.group(1): m.group(2) } ) )
  return info

#
#
# taken from: http://stackoverflow.com/questions/1038824/how-do-i-remove-a-substring-from-the-end-of-a-string-in-python
def strip_end(text, suffix):
    if not text.endswith(suffix):
        return text
    return text[:len(text)-len(suffix)]

# finds valid sd cards inserted in the current system
# currently OSX only as it relies on diskutil
# TODO@Wynter: Add support for Linux and Windows
def find_valid_sd_cards():
  sd_cards = []
  diskutil_output = call_and_return("diskutil", "list")
  disks = find_list_items_containing( "\d\:\s+[a-zA-Z_]+\s+\*(\d+\.\d+\s+[kKgGmM]B).*(disk[0-9])", diskutil_output, (
      lambda m: {"/dev/"+m.group(2): {"name":"/dev/"+m.group(2), "size":m.group(1).strip()} }
      )
    )

  for disk in disks:
    diskutil_output = call_and_return("diskutil", "info", disk)

    if find_list_items_containing("(Protocol):\s+(Secure Digital)", diskutil_output, (lambda m: {"/dev/"+disk: m.group(2)} )):
      sd_cards.append( disks[ disk ] )
    if find_list_items_containing("(Protocol):\s+(USB)", diskutil_output, (lambda m: {"/dev/"+disk: m.group(2)} )):
      sd_cards.append( disks[ disk ] )
  return sd_cards

# finds all fat32 partitions inserted in the system
# you may optionally pass a disk name in to have it
# only find fat partitions on the specified disk
def find_fat_partitions( disk=None ):
  valid_paths = []
  diskutil_args = ["diskutil", "list"]
  if disk:
    diskutil_args.append( disk )

  diskutil_output = call_and_return( *diskutil_args )
  search_regex = "\d\:\s+.*FAT_32\s+(.+)\s+\d+\.\d+\s+[kKgGmM]B+.*(disk[0-9]s[0-9])"
  fat_partitions = find_list_items_containing( search_regex,
    diskutil_output,
    ( lambda m: [m.group(1).strip(), m.group(2)] )
  )
  return fat_partitions


import os
import click
import time
import sys
from .util import Config
from .util import Project
from .util import Util
from .util import OS
from distutils import spawn
from .commands import *

#
# Stak command group
#
@click.group()
@click.option('--verbose', is_flag=True , help="Prints extra information for each command.")
@click.option('--stak-directory', type=click.Path(), help="Location of your .stak directory.")
def stak( verbose, stak_directory):
  """STAK is a collection of tools for building projects on embedded Linux"""

  Config.verbose = verbose
  Config.stak_directory = stak_directory
  Config.local_data = Config.load( )
  Config.load_globals( )
  if not Config.global_data:
    Config.global_data = {}
  Config.current_device = None

  # make sure we have the following directories
  # ~/.stak
  # ~/.stak/images
  Util.create_folders_if_nonexistent( "~/.stak", "~/.stak/images" )


  # test if honcho exists
  if not spawn.find_executable("honcho"):
    click.echo("Could not find honcho executable. Exiting...")
    sys.exit()


#def iterate():
#  number = 0
#  while number < 1000:
#    yield number
#    number+=1
#
#def show_iteration(item):
#  if item:
#    return "Iteration: %d" % item
#
#@stak.command()
#def jump():
#  with click.progressbar(iterable=iterate(), width=10, show_percent=False,
#                          show_eta=False, item_show_func=show_iteration,
#                          fill_char=click.style('#', fg='magenta')) as bar:
#    for item in bar:
#      time.sleep(0.0313)

#
# Create command
#
@stak.command()
@click.argument('project')
@click.option('--language',
              default=None,
              type=click.Choice( Config.example_languages )
  )
def create( project, language ):
  """Creates a new STAK project."""
  Create.create( project, language )

#
# device [command]
#
@stak.group()
def device():
  """Tools for managing hardware running STAK OS"""
  pass


@device.command("setup-card")
def setup_card():
  """Downloads and flashes the STAK OS"""
  Devices.setup_card()
#
# device add [device]
#
@device.command()
@click.argument('device')
def add( device ):
  Devices.add( device )
#
# device remove [device]
#
@device.command()
@click.argument('device')
def remove( device ):
  """Remove device from project"""
  Devices.remove( device )

@device.command("set-default")
@click.argument('device')
def set_default( device ):
  Devices.set_default_device( device )
#
#
# device list
#
@device.command()
def list( ):
  """List all STAK devices on your network"""
  Devices.list( )

#
# device discover
#
#   search the network for any devices broadcasting a stak zeroconf service
#
@device.command()
def discover( ):
  Devices.discover()


@device.command()
@click.argument('name', default=None, required=False)
@click.argument('parameter', type=click.STRING)
@click.argument('value', type=click.STRING)
def set( name, parameter, value ):
  if name:
    Devices.set_property(name, parameter, value)

@device.command()
@click.argument('name', default=None, required=False)
@click.argument('parameter', type=click.STRING)
def get( name, parameter ):
  if name:
    Devices.get_property(name, parameter)


#
# Build command
#
@stak.command()
@click.argument('project',
                required=False,
                default=os.getcwd(),
                type=click.Path()
  )
@click.option('--target',
                required=False,
                default="debug",
                type=click.STRING
  )
def build( project, target ):
  """Cross-compile your project"""
  Project.build( project, target )
  # click.echo("Comming Soon to Terminal Near You...")
#
# Run command
#
@stak.command()
@click.argument('project',
                required=False,
                default=os.getcwd(),
                type=click.Path()
  )
@click.option('--target',
                required=False,
                default="debug",
                type=click.STRING
  )
@click.option('--device',type=click.STRING)
def run( project, target, device ):
  """Run code built with STAK BUILD on device"""
  Project.run( project, target )

#
# ssh command
#
@stak.command()
@click.argument('project',
                required=False,
                default=os.getcwd(),
                type=click.Path()
  )
@click.option('--target',
                required=False,
                default="debug",
                type=click.STRING
  )
@click.option('--device',type=click.STRING)
def ssh( project, target, device ):
  """Open SSH session on device"""
  Project.ssh( project, target )


#
# Deploy command
#
@stak.command()
@click.argument('project',
                required=False,
                default=os.getcwd(),
                type=click.Path()
  )
@click.option('--target',
                required=False,
                default="debug",
                type=click.STRING
  )
@click.option('--device',type=click.STRING)
def deploy( project, target, device ):
  """Installs project as system service that runs on BOOT"""
  Project.deploy( project, target )

#
# Burn command
#
@stak.command()
@click.option('--image', required=False, default=None, type=int)
def burn( image ):
  """Burns the specified OS to an SD card"""
  OS.burn( image )


#!/usr/bin/env python
#
# Copyright (c) 2014, Mike 'Fuzzy' Partin <fuzzy@fu-manchu.org>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL COPYRIGHT HOLDER BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

###########
## Stdlib

import os
import sys
import time
import urllib2
import urlparse
import click
from curses import wrapper

####################
## Utility classes

class HumanReadable:

    def __init__(self):
        print('HumanReadable')

    def size(self, bytes=0):
        kbyte = 1024
        mbyte = (kbyte**2)
        gbyte = (kbyte**3)
        tbyte = (kbyte**4)
        pbyte = (kbyte**5)
        ebyte = (kbyte**6)
        zbyte = (kbyte**7)

        if bytes < kbyte: 
            retv = '%dB' % int(bytes)
            return '%9s' % retv
        elif bytes >= kbyte and bytes < mbyte:
            retv = '%04.02fKB' % (float(bytes) / float(kbyte))
            return '%9s' % retv
        elif bytes >= mbyte and bytes < gbyte:
            retv = '%04.02fMB' % (float(bytes) / float(mbyte))
            return '%9s' % retv
        elif bytes >= gbyte and bytes < tbyte:
            retv = '%04.02fGB' % (float(bytes) / float(gbyte))
            return '%9s' % retv
        elif bytes >= tbyte and bytes < pbyte:
            retv = '%04.02fTB' % (float(bytes) / float(tbyte))
            return '%9s' % retv
        elif bytes >= pbyte and bytes < ebyte:
            retv = '%04.02fPB' % (float(bytes) / float(pbyte))
            return '%9s' % retv
        elif bytes >= ebyte and bytes < zbyte:
            retv = '%04.02fEB' % (float(bytes) / float(ebyte))
            return '%9s' % retv
        else:
            retv = '%04.02fZB' % (float(bytes) / float(zbyte))
            return '%9s' % retv

    def time(self, seconds=0):
        # These are for convenience
        minute = 60
        hour   = (minute**2)
        day    = (hour*24)
        week   = (day*7)
        month  = (week*4)
        year   = (month*12)

        secs, mins, hrs, days, weeks, months, years = 0, 0, 0, 0, 0, 0, 0

        if seconds > year:
            years   = (seconds / year)
            tmp     = (seconds % year)
            seconds = tmp
        if seconds > month:
            months  = (seconds / month)
            tmp     = (seconds % month)
            seconds = tmp
        if seconds > week:
            weeks   = (seconds / week)
            tmp     = (seconds % week)
            seconds = tmp
        if seconds > day:
            days    = (seconds / day)
            tmp     = (seconds % day)
            seconds = tmp
        if seconds > hour:
            hrs     = (seconds / hour)
            tmp     = (seconds % hour)
            seconds = tmp
        if seconds > minute:
            mins    = (seconds / minute)
            secs    = (seconds % minute)
        if seconds < minute:
            secs   = seconds

        if years != 0:
            return '%4dy%2dm%1dw%1dd %02d:%02d:%02d' % (
                years, months, weeks, days, hrs, mins, secs
            )
        if months != 0:
            return '%2dm%1dw%1dd %02d:%02d:%02d' % (
                months, weeks, days, hrs, mins, secs
            )
        if weeks != 0:
            return '%1dw%1dd %02d:%02d:%02d' % (
                weeks, days, hrs, mins, secs
            )
        if days != 0:
            return '%1dd %02d:%02d:%02d' % (days, hrs, mins, secs)
        
        return '%02d:%02d:%02d' % (hrs, mins, secs)

class Output(HumanReadable):
    def __init__(self):
        self.display_flag  = True
        self.display_count = None

        # This is the most portable way (across POSIX systems) to get
        # our screen size that I can find, so, screw windows. Yeah.
        wrapper(self.__setmaxyx)

    def __setmaxyx(self, stdscr):
        (self.max_y, self.max_x) = stdscr.getmaxyx()


    def display(self, inTot=None, inSz=None, outSz=None, start_time=None):
        try:
            a = self.lastupdate
        except AttributeError:
            self.lastupdate = (time.time() - 5)
            
        if inTot:
            remain  = (inTot - outSz)
            percent = (float(outSz) / float(inTot))
            elapsed = (time.time() - start_time)
            speed   = (float(outSz) / float(elapsed))
            eta     = int(remain / speed)
                
            # now build out the majority of the display string
            linest  = '%s in %s @ %s/sec [' % (
                self.size(outSz),
                self.time(elapsed),
                self.size(speed)
            )
            lineend = '] %3d%% eta %s' % (
                int(percent * 100),
                self.time(eta)
            )
            
            # now figure out how many hashmarks we need
            curlen  = (len(linest)+len(lineend))
            hashlen = (self.max_x - curlen)
            hashes  = int(hashlen * percent)
            padding = (hashlen - hashes)
                
            # and put the line together
            line    = '%s%s%s%s' % (
                linest,
                click.style('#',fg="green")*hashes,
                click.style('-',fg="magenta")*padding,
                lineend
            )

        else:
            line = '%s in %s @ %s/sec' % (
                self.size(outSz),
                self.time(time.time() - start_time),
                self.size((outSz / (time.time() - start_time)))
            )

        if (time.time() - self.lastupdate) >= 1 and not self.quiet:
            sys.stderr.write('\r%s' % line)
            sys.stderr.flush()
            self.lastupdate = time.time()

################
## I/O classes

class Io:
    def _validateTarget(self, target=None):
        # First lets see if this is a legit path/file
        if os.path.exists(target):
            return True
        result = urlparse.urlparse(target)
        if result.scheme in ['http', 'https', 'ftp', 'ftps', 'scp', 'sftp', 'file']:
            return True
        return False

    def close(self):
        self.target.close()

    def read(self, bsize=40960):
        return self.target.read(bsize)

    def write(self, data):
        return self.target.write(data)

    def seek(self, pos=None):
        if pos:
            return self.target.seek(pos)                

class FileIo(Io):
    def __init__(self, target=None, mode=None):
        if not self._validateTarget(target):
            mode = 'w+'
        self.target = open(target, mode)
        if self.target.name != '/dev/stdout' and self.target.name != '/dev/stdin':
            self.size  = os.stat(self.target.name).st_size
            self.pipe  = False
        else:
            self.size  = None
            self.pipe  = True

class FifoIo(FileIo):
    pass

class HttpIo(Io):
    def __init__(self, target=None):
        self._validateTarget(target)
        self.target = urllib2.urlopen(target)

class FtpIo(HttpIo):
    pass

class ScpIo(Io):
    def __init__(self, bsize=10240, skip=0):
        pass

###################
## Transfer class

class Transfer(Output):
    def __init__(
            self,
            src=None,
            dst=None,
            tsize=None,
            bsize=40960,
            count=None,
            skip=None,
            quiet=False
    ):
        Output.__init__(self)
        self.src   = src
        self.dst   = dst
        self.bsize = bsize
        self.tsize = tsize
        self.count = count # TODO: Impliment
        self.skip  = skip  # TODO: Impliment
        self.quiet = quiet
        # Lets set our total size for display purposes
        if self.count:
            self.src.size = (self.bsize * self.count)
        # if we have set our total_size via an argument, lets go ahead, and set that
        try:
            if self.tsize:
                self.src.size = self.tsize
            elif not self.tsize and not os.path.isfile(self.src.target.name):
                self.src.size = -1
        except AttributeError:
            self.src.size = int(self.src.target.headers['content-length'])
        except:
            self.src.size = -1

    def start(self):
        blocks = 0
        totsze = 0

        try:
            st     = time.time()
            buff   = self.src.read(self.bsize)
            while buff:
                if totsze <= self.src.size or self.src.size == -1:
                    self.dst.write(buff)
                    blocks += 1
                    totsze += len(buff)
                    if self.src.size != -1:
                        self.display(self.src.size, totsze, totsze, st)
                    else:
                        self.display(None, totsze, totsze, st)
                    buff    = self.src.read(self.bsize)
                else:
                    buff    = None
            time.sleep(0.25)
            self.display(self.src.size, self.src.size, self.src.size, st)
            print
        except KeyboardInterrupt:
            print
            sys.exit(1)


import click
import yaml
import os, sys
import re
import subprocess
import pwd, grp
from stak.util import Util
from stak.util import DD
from stak.util import Config
import zipfile

supported_operating_systems = [
    { "name": "Stak",               "url": "http://stak-images.s3.amazonaws.com/stak-rpi-base.iso.bz2" },
    { "name": "Raspbian",           "url": "http://downloads.raspberrypi.org/raspbian_latest" },
    { "name": "OpenElec",           "url": "http://downloads.raspberrypi.org/openelec_latest" },
    { "name": "Pidora",             "url": "http://downloads.raspberrypi.org/pidora_latest" },
    { "name": "Ubuntu Snappy Core", "url": "http://downloads.raspberrypi.org/ubuntu_latest" },
    { "name": "Raspbmc",            "url": "http://downloads.raspberrypi.org/raspbmc_latest" },
    { "name": "RISC OS",            "url": "http://downloads.raspberrypi.org/riscos_latest" },
]

#
#
#
def get_download_url( os_number = 0 ):
    global supported_operating_systems
    url = supported_operating_systems[ os_number ]["url"]
    info = Util.get_http_header( url )

    if "Location" in info:
        return info["Location"]
    return url

#
#
#
def get_image_filename( os_number = 0 ):
    # handle things differently for Stak OS
    if os_number == 0:
        return Util.strip_end(os.path.basename( supported_operating_systems[ os_number ]["url"] ), ".bz2")
    else:
        return supported_operating_systems[ os_number ]["name"]+".img"

#
#
#
def get_image_compressed_filepath( os_number = 0 ):
    return os.path.expanduser( "~/.stak/images/" + os.path.basename( supported_operating_systems[ os_number ]["url"] ) )

#
#
#
def get_image_filepath( os_number = 0 ):
    # handle things differently for Stak OS
    image_name = get_image_filename( os_number )
    return os.path.expanduser( "~/.stak/images/" + image_name )

#
#
#
def get_image_exists( os_number = 0 ):
    global supported_operating_systems
    
    image_name = get_image_filename( os_number )
    image_path = os.path.expanduser("~/.stak/images/" + image_name)

    return os.path.exists(image_path)

#
#
#
def download_os_image( os_number = 0 ):
    
    image_path = get_image_filepath( os_number )
    image_compressed_filepath = get_image_compressed_filepath( os_number )
    image_url = get_download_url( os_number )

    if not get_image_exists( os_number ):
        if not os.path.exists( image_compressed_filepath ):
            Util.dd( image_url, image_compressed_filepath )
        
        click.echo("Decompressing image...")
        
        if ".bz2" in image_compressed_filepath:
            if os.path.exists( image_path ):
                os.remove( image_path )
            Util.call_and_return("bunzip2", image_compressed_filepath )
        else:
            if os.path.exists( image_path ):
                os.remove( image_path )
            # Util.call_and_return("unzip", image_compressed_filepath )
            fh = open(image_compressed_filepath, 'rb')
            z = zipfile.ZipFile(fh)
            for name in z.namelist():
              click.echo("file: "+ name )
              outfile = open(image_path, 'wb')
              outfile.write( z.read(name) )
              outfile.close()
            fh.close()

#
#
#
def get_latest_image( os_number = 0 ):
    url = get_download_url( os_number )
    info = Util.get_http_header( url )
    image_path = get_image_filepath( os_number )

    # does the image contain ETag information?
    if "ETag" in info:

        # if so, we'll use that as a checksum
        info["ETag"] = info["ETag"].replace("\"", "").rstrip()
        if( "latest-checksum" in Config.global_data and
            Config.global_data["latest-checksum"] == info["ETag"] and
            os.path.exists( image_path ) ):
            return image_path
        else:
            Config.global_data["latest-checksum"] = info["ETag"]
    else:

        # if not, make sure the file exists
        # TODO(Wynter): we need to compare the
        #   filename for date information.
        if os.path.exists( image_path ):
            return image_path

    # we will only get to this point if we need to download a new image
    download_os_image( os_number )

    # we should also save the globals if the new image was created
    if os.path.exists( image_path ):
        Config.save_globals()
    return False

#
#
#
def burn_os( os_number, disk ):
    # we need to unmount all partitions on the specified disk
    # or the os will complain about the disks being in use
    Util.call_and_return( "diskutil","unmountDisk", disk )

    # find the location of the raspbian image.
    # this is located inside of a hidden folder in the user directory for now
    # TODO(Wynter): we should also verify that the file exists and if not download it
    raspbian_image = os.path.expanduser("~/.stak/images/stak-rpi-base.iso")

    # output some information for the user to see that something is going on
    # TODO@Wynter: get some sort of progress output from DD
    click.echo("Writing to "+disk+"...")

    # flash the disk
    rdisk = disk.replace("/dev/disk", "/dev/rdisk")

    if rdisk:
        Util.dd( raspbian_image, rdisk )
    else:
        Util.dd( raspbian_image, disk )

    Util.call_and_return( "diskutil","eject", disk )

# locate and set up an sd card with a new stak os image
def burn( os_number ):

  if not os_number:
    Util.display_list( "Supported OSes", supported_operating_systems, (
        lambda index,value: click.style( value["name"], bold=True, fg="cyan" )
          + ": "
          + (click.style( "Downloaded", fg="green") if get_image_exists(index) else click.style( "Not Downloaded", fg="red" ))
      )
    )
    os_number = click.prompt("Please select the OS to install", type=int)

  get_latest_image( os_number )

  euid = os.geteuid()
  if euid != 0:
    click.secho('Super user privileges are required to flash an sd card.', reverse=True)
    Util.raise_privileges(["--image="+str(os_number)])
   
  inserted=False
  cards = None

  # find and verify that the card is inserted
  while not inserted:
    if not click.confirm('Is the SD card inserted?'):
      click.echo("Please insert the SD card now.")
      click.pause()
    else:
      cards = Util.find_valid_sd_cards()
      if not cards:
        click.echo("No sd card found!")
        click.echo("Please insert the SD card now.")
        click.pause()
      else:
        inserted = True

  # have user select the disk to use
  Util.display_list( "Cards found", cards, (
      lambda index,value: click.style( value["name"], bold=True, fg="cyan" )
        + ": "
        + click.style( value["size"], fg="green" )
    )
  )
  selected_index = click.prompt("Please select the card to flash", type=int)
  selected_card = cards[ selected_index ]["name"]
 
  click.wrap_text("text\x08", width=78, initial_indent='', subsequent_indent='', preserve_paragraphs=True)
  click.secho('WARNING!! Make SURE this is the correct card, this command will delete EVERYTHING on the target volume '+selected_card + " of size " + cards[ selected_index ]["size"], reverse=True)
  if not click.confirm('Are you sure you wish to continue?'):
    return
 
  # flash the selected sd card with the stak os base image
  burn_os( os_number, selected_card )
  Util.drop_privileges()
